#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import unicode_literals

from IPython.core.pylabtools import print_figure
import matplotlib.pyplot as plt
from matplotlib.cm import get_cmap
import itertools
import codecs
import tempfile
import networkx as nx
import numpy as np
import subprocess

from segment import Segment

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


class Notebook:
    pass

_notebook = Notebook()
_notebook.extent = None
_DEFAULT_NOTEBOOK_WIDTH = 20
_notebook.width = _DEFAULT_NOTEBOOK_WIDTH


def set_notebook_crop(segment=None, margin=0.1):

    if segment is None:
        _notebook.extent = None
    else:
        assert isinstance(segment, Segment)
        assert segment
        duration = segment.duration
        _notebook.extent = Segment(
            segment.start - margin * duration,
            segment.end + margin * duration
        )


def set_notebook_width(inches=None):
    if inches is None:
        _notebook.width = _DEFAULT_NOTEBOOK_WIDTH
    else:
        _notebook.width = inches

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def _setup(ylim=None, yaxis=False):
    """Prepare figure"""
    fig, ax = plt.subplots()
    ax.set_xlim(_notebook.extent)
    ax.set_xlabel('Time')
    if ylim is not None:
        ax.set_ylim(ylim)
    ax.axes.get_yaxis().set_visible(yaxis)
    return fig, ax


def _render(fig):
    """Render figure as png and return raw image data"""
    # render figure as png
    data = print_figure(fig, 'png')
    # prevent IPython notebook from displaying the figure
    plt.close(fig)
    # return raw image data
    return data


def _draw_segment(ax, segment, y, color, label=None, text=True,
                  boundaries=True):

    # do nothing if segment is empty
    if not segment:
        return

    # draw segment
    ax.hlines(y, segment.start, segment.end, color, lw=1, label=label)
    if boundaries:
        ax.vlines(segment.start, y + 0.05, y - 0.05, color, lw=1)
        ax.vlines(segment.end, y + 0.05, y - 0.05, color, lw=1)

    if label is None:
        return

    # draw label
    if text:
        ax.text(
            segment.middle,
            y + 0.05,
            codecs.encode(unicode(label), 'ascii', 'replace'),
            horizontalalignment='center',
            fontsize=10
        )

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def repr_segment(segment):
    """Get `png` data for `segment`"""

    # remember current figure size
    figsize = plt.rcParams['figure.figsize']
    # and update it for segment display
    plt.rcParams['figure.figsize'] = (_notebook.width, 1)

    if not _notebook.extent:
        set_notebook_crop(segment=segment)

    fig, ax = _setup(ylim=(0, 1))

    # segments are vertically centered and blue
    y = 0.5
    color = 'b'
    _draw_segment(ax, segment, y, color)

    data = _render(fig)

    # go back to previous figure size
    plt.rcParams['figure.figsize'] = figsize

    return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def _y(segments):
    """

    Parameters
    ----------
    segments : iterator
        `Segment` iterator (sorted)

    Returns
    -------
    y : np.array
        y coordinates of each segment

    """

    # up_to stores the largest end time
    # displayed in each line (at the current iteration)
    # (at the beginning, there is only one empty line)
    up_to = [-np.inf]

    # y[k] indicates on which line to display kth segment
    y = []

    for segment in segments:
        # so far, we do not know which line to use
        found = False
        # try each line until we find one that is ok
        for i, u in enumerate(up_to):
            # if segment starts after the previous one
            # on the same line, then we add it to the line
            if segment.start >= u:
                found = True
                y.append(i)
                up_to[i] = segment.end
                break
        # in case we went out of lines, create a new one
        if not found:
            y.append(len(up_to))
            up_to.append(segment.end)

    # from line numbers to actual y coordinates
    y = 1. - 1. / (len(up_to) + 1) * (1 + np.array(y))

    return y

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def repr_timeline(timeline):
    """Get `png` data for `timeline`"""

    # remember current figure size
    figsize = plt.rcParams['figure.figsize']
    # and update it for segment display
    plt.rcParams['figure.figsize'] = (_notebook.width, 1)

    if not _notebook.extent and timeline:
        set_notebook_crop(segment=timeline.extent())

    cropped = timeline.crop(_notebook.extent, mode='loose')

    fig, ax = _setup(ylim=(0, 1))

    color = 'b'
    for segment, y in itertools.izip(cropped, _y(cropped)):
        _draw_segment(ax, segment, y, color)

    data = _render(fig)

    # go back to previous figure size
    plt.rcParams['figure.figsize'] = figsize

    return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def repr_annotation(annotation):
    """Get `png` data for `annotation`"""

    # remember current figure size
    figsize = plt.rcParams['figure.figsize']
    # and update it for segment display
    plt.rcParams['figure.figsize'] = (_notebook.width, 2)

    if not _notebook.extent:
        set_notebook_crop(segment=annotation.get_timeline().extent())

    cropped = annotation.crop(_notebook.extent, mode='intersection')
    segments = [s for s, _ in cropped.itertracks()]

    # one color per label
    chart = cropped.chart()
    cm = get_cmap('gist_rainbow')
    colors = {
        label: cm(1. * i / len(chart))
        for i, (label, _) in enumerate(chart)
    }

    fig, ax = _setup(ylim=(0, 1))

    for (segment, track, label), y in itertools.izip(
            cropped.itertracks(label=True), _y(segments)):
        color = colors[label]  # color = f(label)
        _draw_segment(ax, segment, y, color, label=label)

    data = _render(fig)

    # go back to previous figure size
    plt.rcParams['figure.figsize'] = figsize

    return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def repr_scores(scores):
    """Get `png` data for `scores`"""

    # remember current figure size
    figsize = plt.rcParams['figure.figsize']
    # and update it for segment display
    plt.rcParams['figure.figsize'] = (_notebook.width, 5)

    if not _notebook.extent:
        set_notebook_crop(segment=scores.get_timeline().extent())

    cropped = scores.crop(_notebook.extent, mode='loose')

    cm = get_cmap('gist_rainbow')
    labels = sorted(scores.labels())
    colors = {label: cm(1. * i / len(labels))
              for i, label in enumerate(labels)}

    m = scores._df.min().min()
    M = scores._df.max().max()
    ylim = (m - 0.1 * (M - m), M + 0.1 * (M - m))

    fig, ax = _setup(yaxis=True, ylim=ylim)

    for segment, track, label, value in cropped.itervalues():
        color = colors[label]
        y = value
        _draw_segment(ax, segment, y, color, label=label, text=False,
                      boundaries=False)

    # for segment, track, label, value in cropped.nbest(1).itervalues():
    #     color = colors[label]
    #     y = value
    #     _draw_segment(ax, segment, y, color, label=label, text=True,
    #                   boundaries=True)

    # get one handle per label and plot the corresponding legend
    H, L = ax.get_legend_handles_labels()
    handles = {}
    for h, l in zip(H, L):
        if l in labels:
            handles[l] = h
    ax.legend([handles[l] for l in sorted(handles)], sorted(handles))

    data = _render(fig)

    # go back to previous figure size
    plt.rcParams['figure.figsize'] = figsize

    return data

# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


def _shorten_long_text(text, max_length=30):
    suffix = "..."
    if len(text) > max_length:
        return text[:max_length - len(suffix)] + suffix
    else:
        return text


def _remove_non_ascii(text):
    # list of characters to remove
    # those characters make resulting SVG invalid
    remove = [u"&", ]
    remove = {ord(r): u"" for r in remove}
    ascii = unicode(codecs.encode(text, 'ascii', 'replace'))
    return ascii.translate(remove)


def _dottable(transcription):
    # create new graph with graphviz/dot layout instructions
    # (e.g. graph orientation, node and edge labels, etc.)

    # create a dumb networkx copy to avoid messing with input transcription
    dottable = transcription.copy()

    # graphviz/dot will display graph from left to right
    dottable.graph['graph'] = {'rankdir': 'LR'}

    # set shape, label and tooltip of each node
    for n in transcription.nodes_iter():

        dottable.node[n] = {
            'label': str(n), 'tooltip': str(n),
            'shape': 'circle' if n.drifting else 'box'
        }
        # 'URL': 'javascript:console.log("{t}")'.format(t=n.T),

    # set edge label

    label_header = (
        "<<table border='0' cellborder='0' cellspacing='0' cellpadding='3'>"
    )
    label_pattern = (
        "<tr>"
        "<td align='left'><b>{name}</b></td>"
        "<td align='left'>{value}</td>"
        "</tr>"
    )
    label_footer = "</table>>"
    tooltip_pattern = "[{name}] {value}"

    for source, target, key, data in transcription.edges_iter(
        keys=True, data=True
    ):
        tooltip = ""
        label = ""
        if data:

            # initialize label table
            label = label_header

            for name, value in data.iteritems():
                # remove non-ascii characters
                name = _remove_non_ascii(name)
                value = _remove_non_ascii(value)
                # shorten long value
                short_value = _shorten_long_text(value)
                # update label and tooltip
                label += label_pattern.format(name=name, value=short_value)
                tooltip += tooltip_pattern.format(name=name, value=value)

            # close label table
            label += label_footer

        dottable[source][target][key] = {
            'label': label,
            'labeltooltip': tooltip,
            'edgetooltip': tooltip,
            'headtooltip': tooltip,
            'tailtooltip': tooltip,
        }

    return dottable


def _write_temporary_dot_file(transcription):
    _, path = tempfile.mkstemp('.dot')
    nx.write_dot(_dottable(transcription), path)
    return path


def repr_transcription(transcription):
    """Get `svg` data for `transcription`"""
    path = _write_temporary_dot_file(transcription)
    data = subprocess.check_output(["dot", "-T", "svg", path])
    return data[data.find("<svg"):]



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import unicode_literals

import numpy as np
import pandas
from segment import SEGMENT_PRECISION

class LabelMatrix(object):

    def __init__(self, data=None, dtype=None, rows=None, columns=None):
        super(LabelMatrix, self).__init__()
        if data is None and dtype is None:
            dtype = np.float
        self.df = pandas.DataFrame(
            data=data, dtype=dtype, index=rows, columns=columns)

    def __setitem__(self, (row, col), value):
        self.df = self.df.set_value(row, col, value)
        return self

    def __getitem__(self, (row, col)):
        return self.df.at[row, col]

    def get_rows(self):
        return list(self.df.index)

    def get_columns(self):
        return list(self.df.columns)

    def __get_shape(self):
        return self.df.shape
    shape = property(fget=__get_shape)

    def __nonzero__(self):
        N, M = self.df.shape
        return N*M != 0

    def itervalues(self):
        for row in self.get_rows():
            for col in self.get_columns():
                val = self.df.at[row, col]
                if not np.isnan(val):
                    yield row, col, val

    def argmax(self, axis=None):
        """
        Labels of the maximum values along an axis.

        Parameters
        ----------
        axis : int, optional
            By default, labels are into the whole matrix, otherwise
            along the specified axis (rows or columns)

        Returns
        -------
        label_dict : dictionary of labels
            Dictionary of labels into the matrix.
            {col_label : max_row_label} if axis == 0
            {row_label : max_col_label} if axis == 1
            {max_row_label : max_col_label} if axis == None
        """

        if axis == 0:
            return {c: r
                    for (c, r) in self.df.idxmax(axis=axis).iteritems()}

        elif axis == 1:
            return {r: c
                    for (r, c) in self.df.idxmax(axis=axis).iteritems()}

        else:
            values = [
                (_r, _c, self.df.loc[_r, _c])
                for (_c, _r) in self.df.idxmax(axis=0).iteritems()
            ]
            r, c, _ = sorted(values, key=lambda v: v[2])[-1]
            return {r: c}

    def __neg__(self):
        negated = LabelMatrix()
        negated.df = -self.df
        return negated

    def argmin(self, axis=None):
        """
        Labels of the minimum values along an axis.

        Parameters
        ----------
        axis : int, optional
            By default, labels are into the whole matrix, otherwise
            along the specified axis (rows or columns)

        Returns
        -------
        label_dict : dictionary of labels
            Dictionary of labels into the matrix.
            {col_label : max_row_label} if axis == 0
            {row_label : max_col_label} if axis == 1
            {max_row_label : max_col_label} if axis == None
        """

        return (-self).argmax(axis=axis)

    def __get_T(self):
        transposed = LabelMatrix()
        transposed.df = self.df.T
        return transposed
    T = property(fget=__get_T)

    def remove_column(self, col):
        del self.df[col]
        return self

    def remove_row(self, row):
        df = self.df.T
        del df[row]
        self.df = df.T
        return self

    def copy(self):
        copied = LabelMatrix()
        copied.df = self.df.copy()
        return copied

    def subset(self, rows=None, columns=None):

        if rows is None:
            rows = set(self.get_rows())

        if columns is None:
            columns = set(self.get_columns())

        remove_rows = set(self.get_rows()) - rows
        remove_columns = set(self.get_columns()) - columns

        copied = self.copy()
        for row in remove_rows:
            copied = copied.remove_row(row)
        for col in remove_columns:
            copied = copied.remove_column(col)

        return copied

    def __gt__(self, value):
        compared = LabelMatrix()
        compared.df = self.df > value
        return compared

    def __str__(self):
        return str(self.df)

    def _repr_html_(self):
        return self.df._repr_html_()


def get_cooccurrence_matrix(R, C):

    # initialize label matrix with zeros
    rows = R.labels()
    cols = C.labels()
    K = np.zeros((len(rows), len(cols)), dtype=np.float)
    M = LabelMatrix(data=K, rows=rows, columns=cols)

    # loop on intersecting tracks
    for (r_segment, r_track), (c_segment, c_track) in R.co_iter(C):
        # increment
        r_label = R[r_segment, r_track]
        c_label = C[c_segment, c_track]
        duration = (r_segment & c_segment).duration
        M[r_label, c_label] += duration

    return M



def get_tfidf_matrix(words, documents, idf=True, log=False):
    """Term Frequency Inverse Document Frequency (TF-IDF) confusion matrix

    C[i, j] = TF(i, j) x IDF(i) where
        - documents are J labels
        - words are co-occurring I labels

                  duration of word i in document j         confusion[i, j]
    TF(i, j) = --------------------------------------- = -------------------
               total duration of I words in document j   sum confusion[:, j]

                        number of J documents
    IDF(i) = ----------------------------------------------
             number of J documents co-occurring with word i

                      Nj
           = -----------------------
             sum confusion[i, :] > 0

    Parameters
    ---------
    words : :class:`pyannote.base.annotation.Annotation`
        Every label occurrence is considered a word
        (weighted by the duration of the segment)
    documents : :class:`pyannote.base.annotation.Annotation`
        Every label is considered a document.
    idf : bool, optional
        If `idf` is set to True, returns TF x IDF.
        Otherwise, returns TF. Default is True
    log : bool, optional
        If `log` is True, returns TF x log IDF
    """
    M = get_cooccurrence_matrix(words, documents)
    Nw, Nd = M.shape

    if Nd == 0:
        return M

    K = M.df.values
    rows = M.get_rows()
    cols = M.get_columns()

    # total duration of all words cooccurring with each document
    # np.sum(self.M, axis=0)[j] = 0 ==> self.M[i, j] = 0 for all i
    # so we can safely use np.maximum(1e-3, ...) to avoid DivideByZero
    tf = K / np.tile(
        np.maximum(SEGMENT_PRECISION, np.sum(K, axis=0)),
        (Nw, 1)
    )

    # use IDF only if requested (default is True ==> use IDF)
    if idf:

        # number of documents cooccurring with each word
        # np.sum(self.M > 0, axis=1)[i] = 0 ==> tf[i, j] = 0 for all i
        # and therefore tf.idf [i, j ] = 0
        # so we can safely use np.maximum(1, ...) to avoid DivideByZero
        idf = np.tile(
            float(Nd) / np.maximum(1, np.sum(K > 0, axis=1)),
            (Nd, 1)).T

        # use log only if requested (defaults is False ==> do not use log)
        if log:
            idf = np.log(idf)

    else:
        idf = 1.

    return LabelMatrix(data=tf * idf, rows=rows, columns=cols)

if __name__ == "__main__":
    import doctest
    doctest.testmod()



#!/usr/bin/env python
# encoding: utf-8

# The MIT License (MIT)

# Copyright (c) 2014 CNRS (Hervé BREDIN - http://herve.niderb.fr)

# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:

# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.

# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

from __future__ import unicode_literals

import simplejson as json

PYANNOTE_JSON_SEGMENT = 'S'
PYANNOTE_JSON_TIMELINE = 'L'
PYANNOTE_JSON_ANNOTATION = 'A'
PYANNOTE_JSON_TRANSCRIPTION = 'T'

def object_hook(d):
    """
    Usage
    -----
    >>> import simplejson as json
    >>> with open('file.json', 'r') as f:
    ...   json.load(f, object_hook=object_hook)
    """

    from segment import Segment
    from timeline import Timeline
    from annotation import Annotation
    from transcription import Transcription

    if PYANNOTE_JSON_SEGMENT in d:
        return Segment.from_json(d)

    if PYANNOTE_JSON_TIMELINE in d:
        return Timeline.from_json(d)

    if PYANNOTE_JSON_ANNOTATION in d:
        return Annotation.from_json(d)

    if PYANNOTE_JSON_TRANSCRIPTION in d:
        return Transcription.from_json(d)

    return d

def load(path):
    with open(path, 'r') as f:
        data = json.load(f, encoding='utf-8', object_hook=object_hook)
    return data

def dump(data, path):
    # TODO: add pyannote.core version
    with open(path, 'w') as f:
        json.dump(data, f, encoding='utf-8', for_json=True)


